"""Copy / Paste functionality."""

from __future__ import annotations

from collections.abc import Collection

from gi.repository import Gdk, GLib, GObject

from gaphor.core import Transaction
from gaphor.core.modeling import Presentation
from gaphor.diagram.copypaste import copy_full, paste_full, paste_link


class CopyBuffer(GObject.Object):
    def __init__(self, buffer):
        super().__init__()
        self.buffer = buffer

    buffer = GObject.Property(type=object)


class Clipboard:
    """Copy/Cut/Paste functionality for diagrams."""

    def __init__(self, event_manager, element_factory, clipboard=None):
        self.event_manager = event_manager
        self.element_factory = element_factory

        self.clipboard = clipboard or Gdk.Display.get_default().get_clipboard()

    def copy(self, view):
        if items := view.selection.selected_items:
            self._copy(items)

    def cut(self, view):
        items = view.selection.selected_items
        if not items:
            return

        self._copy(items)

        with Transaction(self.event_manager):
            for i in list(items):
                i.unlink()

    def paste_link(self, view):
        self._paste(view, paste_link)

    def paste_full(self, view):
        self._paste(view, paste_full)

    def _copy(self, items: Collection[Presentation]) -> None:
        if items:
            copy_buffer = copy_full(items, self.element_factory.lookup)
            v = GObject.Value(CopyBuffer.__gtype__, CopyBuffer(buffer=copy_buffer))
            self.clipboard.set_content(Gdk.ContentProvider.new_for_value(v))

    def _paste(self, view, paster):
        diagram = view.model

        def on_paste(_source_object, result):
            try:
                copy_buffer = self.clipboard.read_value_finish(result)
            except GLib.GError as e:
                if str(e).startswith("g-io-error-quark:"):
                    return
                raise

            with Transaction(self.event_manager):
                # Create new id's that have to be used to create the items:
                new_items = paster(copy_buffer.buffer, diagram)

                # move pasted items a bit, so user can see result of his action :)
                for item in new_items:
                    if item.parent not in new_items:
                        item.matrix.translate(10, 10)

            selection = view.selection
            selection.unselect_all()
            selection.select_items(*new_items)

        self.clipboard.read_value_async(
            CopyBuffer.__gtype__,
            io_priority=GLib.PRIORITY_DEFAULT,
            cancellable=None,
            callback=on_paste,
        )



import logging
import sys

from gaphas.model import Model
from gaphas.view import GtkView
from gi.repository import GObject, Gtk

from gaphor.diagram.selection import Selection

log = logging.getLogger(__name__)


class DiagramView(GtkView):
    __gtype_name__ = "DiagramView"

    def __init__(self, model: Model | None = None):
        super().__init__(model, Selection())

    @GObject.Signal(name="cut-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _cut_clipboard(self):
        pass

    @GObject.Signal(name="copy-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _copy_clipboard(self):
        pass

    @GObject.Signal(name="paste-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _paste_clipboard(self):
        pass

    @GObject.Signal(name="paste-full-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _paste_full_clipboard(self):
        pass

    @GObject.Signal(name="delete", flags=GObject.SignalFlags.RUN_LAST)
    def _delete(self):
        pass


def _trigger_signal(signal_name):
    def _trigger_action(self, _action_name, _param):
        self.emit(signal_name)

    return _trigger_action


if hasattr(DiagramView, "install_action"):
    # Deal with Gtk being mocked when generating docs
    DiagramView.install_action("clipboard.cut", None, _trigger_signal("cut-clipboard"))
    DiagramView.install_action(
        "clipboard.copy", None, _trigger_signal("copy-clipboard")
    )
    DiagramView.install_action(
        "clipboard.paste", None, _trigger_signal("paste-clipboard")
    )
    DiagramView.install_action(
        "clipboard.paste-full", None, _trigger_signal("paste-full-clipboard")
    )
    DiagramView.install_action("diagram.delete", None, _trigger_signal("delete"))

    def _new_named_shortcut(shortcut, action_name):
        return Gtk.Shortcut.new(
            trigger=Gtk.ShortcutTrigger.parse_string(shortcut),
            action=Gtk.NamedAction.new(action_name),
        )

    _mod = "<Meta>" if sys.platform == "darwin" else "<Control>"

    DiagramView.add_shortcut(_new_named_shortcut(f"{_mod}x", "clipboard.cut"))
    DiagramView.add_shortcut(_new_named_shortcut(f"{_mod}c", "clipboard.copy"))
    DiagramView.add_shortcut(_new_named_shortcut(f"{_mod}v", "clipboard.paste"))
    DiagramView.add_shortcut(
        _new_named_shortcut(f"{_mod}<Shift>v", "clipboard.paste-full")
    )
    DiagramView.add_shortcut(
        _new_named_shortcut("Delete|BackSpace|<Meta>BackSpace", "diagram.delete")
    )
    del _mod



from gaphor.diagram.support import represents
from gaphor.SysML.sysml import (
    ActivityDiagram,
    BlockDefinitionDiagram,
    InternalBlockDiagram,
    PackageDiagram,
    RequirementDiagram,
    SequenceDiagram,
    StateMachineDiagram,
    UseCaseDiagram,
)
from gaphor.UML.general.diagramitem import DiagramItem

for _type in (
    ActivityDiagram,
    BlockDefinitionDiagram,
    InternalBlockDiagram,
    PackageDiagram,
    RequirementDiagram,
    SequenceDiagram,
    StateMachineDiagram,
    UseCaseDiagram,
):
    represents(_type)(DiagramItem)



"""AsyncIO support for Gaphor.

This module allows to keep track of
async tasks.
"""

import asyncio
import contextlib
from collections.abc import Coroutine

from gi.events import GLibEventLoopPolicy
from gi.repository import GLib


class TaskOwner:
    """Mixin that allows an object to manage an asyncio task.

    The TaskOwner allows for managing one (!) task at a time.
    """

    def __init__(self):
        self._background_task: asyncio.Task | None = None

    def create_background_task(self, coro: Coroutine) -> asyncio.Task:
        assert self._background_task is None or self._background_task.done()
        task = asyncio.create_task(coro)
        self._background_task = task

        def task_done(task):
            assert self._background_task is task
            self._background_task = None

        task.add_done_callback(task_done)
        return task

    def cancel_background_task(self) -> bool:
        return (self._background_task is not None) and self._background_task.cancel()

    async def gather_background_task(self):
        if self._background_task:
            await asyncio.gather(self._background_task)


async def response_from_adwaita_dialog(dialog, window) -> str:
    response = asyncio.get_running_loop().create_future()

    def response_cb(_dialog, answer):
        response.set_result(answer)

    dialog.connect("response", response_cb)
    dialog.present(window)

    answer = await response
    assert isinstance(answer, str)

    return answer


# Notes for PyGObject 3.52:
# * `glib_event_loop_policy` can be removed: use `with GLibEventLoopPolicy()`.
# * `sleep` can be removed: use `task.set_priority(GLib.PRIORITY_LOW)`.


@contextlib.contextmanager
def glib_event_loop_policy():
    original = asyncio.get_event_loop_policy()
    policy = GLibEventLoopPolicy()
    asyncio.set_event_loop_policy(policy)
    try:
        yield policy
    finally:
        asyncio.set_event_loop_policy(original)


def sleep(delay, result=None, priority=GLib.PRIORITY_LOW):
    """Prioritizable version of asyncio.sleep().

    Normal asyncio.sleep() runs with priority DEFAULT,
    which blocks the UI. This sleep has a configurable
    priority.
    """
    if GLib.main_depth() == 0:
        return asyncio.sleep(delay, result)

    f: asyncio.Future = asyncio.Future()

    def timeout_func(*args):
        f.set_result(result)
        return GLib.SOURCE_REMOVE

    GLib.timeout_add(int(delay * 1000), timeout_func, f, priority=priority)
    return f



from gaphor.diagram.support import represents
from gaphor.RAAML.raaml import FTADiagram, STPADiagram
from gaphor.UML.general.diagramitem import DiagramItem

represents(FTADiagram)(DiagramItem)
represents(STPADiagram)(DiagramItem)



"""A generic way to handle errors in GUI applications.

This module also contains a ErrorHandlerAspect, which can be easily
attached to a class method and will raise the error dialog when the
method exits with an exception.
"""

import pdb
import sys

from gi.repository import Adw

from gaphor.asyncio import response_from_adwaita_dialog
from gaphor.i18n import gettext


async def error_dialog(message, secondary_message="", window=None):
    _exc_type, _exc_value, exc_traceback = sys.exc_info()
    atty = __debug__ and exc_traceback and sys.stdin.isatty()

    debug_body = (f"{secondary_message}\n\n" if secondary_message else "") + gettext(
        "It looks like Gaphor is started from the command line. Do you want to open a debug session?"
    )
    dialog = Adw.AlertDialog.new(
        message,
        debug_body if atty else secondary_message,
    )
    dialog.add_response("close", gettext("Close"))
    if atty:
        dialog.add_response("debug", gettext("Start Debug Session"))
        dialog.set_default_response("debug")
    else:
        dialog.set_default_response("close")
    dialog.set_close_response("close")

    answer = await response_from_adwaita_dialog(dialog, window)

    if exc_traceback and answer in (100, "debug"):
        pdb.post_mortem(exc_traceback)



import logging
import os.path
from pathlib import Path

from gi.repository import Adw

from gaphor.abc import ActionProvider, Service
from gaphor.asyncio import TaskOwner
from gaphor.core import action, gettext
from gaphor.ui.filedialog import GAPHOR_FILTER, open_file_dialog

log = logging.getLogger(__name__)


class AppFileManager(Service, ActionProvider, TaskOwner):
    """Handle application level file loading."""

    def __init__(self, application):
        super().__init__()
        self.application = application
        self.last_dir = None

    def shutdown(self):
        self.cancel_background_task()

    @property
    def window(self):
        return self.application.active_window

    @action(name="app.file-open")
    def action_open(self):
        """This menu action opens the standard model open dialog."""

        async def open_files():
            filenames: list[Path] | None = await open_file_dialog(  # type: ignore[assignment]
                gettext("Open a Model"),
                parent=self.window,
                dirname=self.last_dir,
                filters=GAPHOR_FILTER,
            )
            if not filenames:
                return

            for filename in filenames:
                if any(
                    session
                    for session in self.application.sessions
                    if session.filename == filename
                ):
                    name = Path(filename).name
                    title = gettext("Switch to {name}?").format(name=name)
                    body = gettext(
                        "{name} is already opened. Do you want to switch to the opened window instead?"
                    ).format(name=name)

                    # Should show only one dialog at a time

                    dialog = Adw.MessageDialog.new(
                        self.window,
                        title,
                    )
                    dialog.set_body(body)
                    dialog.add_response("open", gettext("Open Again"))
                    dialog.add_response("switch", gettext("Switch"))
                    dialog.set_response_appearance(
                        "switch", Adw.ResponseAppearance.SUGGESTED
                    )
                    dialog.set_default_response("switch")
                    dialog.set_close_response("open")
                    dialog.present()
                    answer = await dialog.choose()
                    dialog.destroy()
                    self.application.new_session(
                        filename=filename, force=(answer == "open")
                    )
                else:
                    self.application.new_session(filename=filename)

                self.last_dir = os.path.dirname(filename)

        self.create_background_task(open_files())



"""State items property pages.

To register property pages implemented in this module, it is imported in
gaphor.adapter package.
"""

from __future__ import annotations

from gi.repository import Gio

from gaphor import UML
from gaphor.diagram.propertypages import (
    LabelValue,
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML.states.state import StateItem
from gaphor.UML.states.statemachine import StateMachineItem

new_builder = new_resource_builder("gaphor.UML.states")


@PropertyPages.register(UML.Transition)
class TransitionPropertyPage(PropertyPageBase):
    """Transition property page allows to edit trigger, guard, and action specification."""

    order = 15

    subject: UML.Transition

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject.watcher()

    def construct(self):
        subject = self.subject
        if not subject:
            return

        builder = new_builder("transition-editor")

        guard = builder.get_object("guard")
        if subject.guard:
            guard.set_text(subject.guard.specification or "")

        @handler_blocking(guard, "changed", self._on_guard_change)
        def guard_handler(event):
            if event.element is subject.guard and guard.get_text() != event.new_value:
                guard.set_text(event.new_value or "")

        self.watcher.watch("guard[Constraint].specification", guard_handler)

        trigger = builder.get_object("trigger")
        if subject.trigger:
            trigger.set_text(subject.trigger.name or "")

        @handler_blocking(trigger, "changed", self._on_trigger_change)
        def trigger_handler(event):
            if (
                event.element is subject.trigger
                and trigger.get_text() != event.new_value
            ):
                trigger.set_text(event.new_value or "")

        self.watcher.watch("trigger[Behavior].name", trigger_handler)

        action = builder.get_object("action")
        if subject.action:
            action.set_text(subject.action.name or "")

        @handler_blocking(action, "changed", self._on_action_change)
        def action_handler(event):
            if event.element is subject.action and action.get_text() != event.new_value:
                action.set_text(event.new_value or "")

        self.watcher.watch("action[Behavior].name", action_handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("transition-editor"), self.watcher
        )

    def _on_guard_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager):
            if not self.subject.guard:
                self.subject.guard = self.subject.model.create(UML.Constraint)
            self.subject.guard.specification = value

    def _on_trigger_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager):
            if not self.subject.trigger:
                self.subject.trigger = self.subject.model.create(UML.Behavior)
            self.subject.trigger.name = value

    def _on_action_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager):
            if not self.subject.action:
                self.subject.action = self.subject.model.create(UML.Behavior)
            self.subject.action.name = value


@PropertyPages.register(UML.State)
class StatePropertyPage(PropertyPageBase):
    """State property page."""

    order = 15
    subject: UML.State

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        subject = self.subject
        if not subject:
            return

        builder = new_builder("state-editor")
        options = self._behavior_options()

        entry_dropdown = builder.get_object("entry")
        entry_dropdown.set_model(options)

        if self.subject.entry:
            entry_dropdown.set_selected(
                next(
                    n
                    for n, lv in enumerate(options)
                    if lv.value == self.subject.entry.id
                )
            )

        entry_dropdown.connect("notify::selected", self._on_entry_behavior_changed)

        exit_dropdown = builder.get_object("exit")
        exit_dropdown.set_model(options)

        if self.subject.exit:
            exit_dropdown.set_selected(
                next(
                    n
                    for n, lv in enumerate(options)
                    if lv.value == self.subject.exit.id
                )
            )

        exit_dropdown.connect("notify::selected", self._on_exit_behavior_changed)

        do_activity_dropdown = builder.get_object("do-activity")
        do_activity_dropdown.set_model(options)

        if self.subject.doActivity:
            do_activity_dropdown.set_selected(
                next(
                    n
                    for n, lv in enumerate(options)
                    if lv.value == self.subject.doActivity.id
                )
            )

        do_activity_dropdown.connect(
            "notify::selected", self._on_do_activity_behavior_changed
        )

        return builder.get_object("state-editor")

    def _behavior_options(self):
        options = Gio.ListStore.new(LabelValue)
        options.append(LabelValue("", None))

        for c in sorted(
            (c for c in self.subject.model.select(UML.Behavior) if c.name),
            key=lambda c: c.name or "",
        ):
            options.append(LabelValue(c.name, c.id))

        return options

    def _on_entry_behavior_changed(self, dropdown, _pspec):
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = self.subject.model.lookup(id)
                assert isinstance(element, UML.Behavior)
                self.subject.entry = element
            else:
                del self.subject.entry

    def _on_exit_behavior_changed(self, dropdown, _pspec):
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = self.subject.model.lookup(id)
                assert isinstance(element, UML.Behavior)
                self.subject.exit = element
            else:
                del self.subject.exit

    def _on_do_activity_behavior_changed(self, dropdown, _pspec):
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = self.subject.model.lookup(id)
                assert isinstance(element, UML.Behavior)
                self.subject.doActivity = element
            else:
                del self.subject.doActivity


@PropertyPages.register(StateItem)
@PropertyPages.register(StateMachineItem)
class RegionPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, item: StateItem | StateMachineItem, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "region-editor",
            "num-regions-adjustment",
            signals={
                "show-regions-changed": (self._on_show_regions_changed,),
                "regions-changed": (self._on_num_regions_changed,),
            },
        )

        num_regions = builder.get_object("num-regions")
        num_regions.set_value(len(self.item.subject.region))

        show_regions = builder.get_object("show-regions")
        show_regions.set_active(self.item.show_regions)

        return builder.get_object("region-editor")

    def _on_num_regions_changed(self, spin_button):
        num_regions = spin_button.get_value_as_int()
        with Transaction(self.event_manager):
            self.update_regions(num_regions)

    def _on_show_regions_changed(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_regions = button.get_active()

    def update_regions(self, num_regions) -> None:
        if not self.item.subject:
            return

        while num_regions > len(self.item.subject.region):
            region = self.item.subject.model.create(UML.Region)
            self.item.subject.region = region

        while num_regions < len(self.item.subject.region):
            region = self.item.subject.region[-1]
            region.unlink()
